//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
#include <vcl\registry.hpp>
#include <sys/stat.h>
#include <dir.h>
#include "ServerMe.h"
#include "MainUnit.h"
#include "AboutBoxUnit.h"
#include "OptionsBoxUnit.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
AnsiString	GBasePath;
bool		GAllowDirectoryListing = true;
bool		GAutoStart = true;
bool		GHideMainWindow = false;
bool		GAllowContentDLLs = false;
bool		GCanExit = false;
//---------------------------------------------------------------------------
TMain *Main;
CWebServer server;
static AnsiString HTMLEscape(AnsiString& s)
{
	AnsiString r;
    for (int i=1; i <= s.Length(); ++i)
    {
    	switch (s[i])
        {
        case '&':
        	r += "&amp;";
            break;
        case '<':
        	r += "&lt;";
            break;
        default:
        	r += s[i];
        }
    }
    return r;
}
static AnsiString URLEscape(AnsiString& s)
{
	const char hex[16] = "0123456789ABCDEF";
	AnsiString r;
    for (int i=1; i <= s.Length(); ++i)
    {
    	if (s[i] < 33)
        {
        	r += '%';
            r += hex[s[i]>>4];
            r += hex[s[i]&7];
        }
        else
        {
        	r += s[i];
        }
    }
    return r;
}
static int HEXNibble(const char n)
{
	return	n >= '0' && n <= '9' ? n - '0' :
            n >= 'a' && n <= 'f' ? n - 'a' :
            n >= 'A' && n <= 'F' ? n - 'A' :
            '_';
}
static int HEXByte(const char* base)
{
	return (HEXNibble(base[0]) << 4)|HEXNibble(base[1]);
}
static AnsiString URLUnescape(AnsiString& s)
{
	AnsiString r;
    for (int i=1; i <= s.Length(); ++i)
    {
    	if (s[i] == '%' && i + 2 <= s.Length())
        {
        	r += (char)HEXByte(s.c_str() + i);
            i += 2;
        }
        else if (s[i] == '+')
        {
        	r += ' ';
        }
        else
        {
        	r += s[i];
        }
    }
    return r;
}
static int _stdcall DLLSend(SOCKET socket, const char* data, int count)
{
	CWebServerSocket sock(socket);
    return sock.Send(data, count) ? 1 : 0;
}

static void Callback(SWebServerRequest& request, void*)
{
	// Log the request
	TListItem* item = Main->lvLog->Items->Add();
    item->Caption = DateTimeToStr(Now());
    item->SubItems->Add(request.Address);
    item->MakeVisible(false);

    // Report
    Main->sbStatus->Panels->Items[1]->Text = "Serving " + AnsiString(request.Address);

    // Try to resolve the requested file
    AnsiString address = request.Address;
    AnsiString filePart;
    if (address.Pos('?') != 0)
    {
    	filePart = URLUnescape(address.SubString(1, address.Pos('?') - 1));
    }
    else
    {
    	filePart = URLUnescape(address);
    }

    // Try to find absolute path
    AnsiString fullFilePath = GBasePath + filePart;
    for (int i=1; i <= fullFilePath.Length(); ++i)
    {
    	if (fullFilePath[i] == '/')
        {
        	fullFilePath[i] = '\\';
        }
    }
    if (fullFilePath[fullFilePath.Length()] == '\\')
    {
    	fullFilePath = fullFilePath.SubString(1, fullFilePath.Length() - 1);
    }
    char fullPath[MAX_PATH];
    if (!GetFullPathName(fullFilePath.c_str(), MAX_PATH, fullPath, 0))
    {
    	request.Socket.SendContent("invalid request");
    	return;
    }
    fullFilePath = fullPath;

    // Check if the path is inside the base path
    if (fullFilePath != GBasePath &&
        fullFilePath.SubString(1, GBasePath.Length() + 1) != GBasePath + '\\')
    {
    	request.Socket.SendContent("invalid path");
        return;
    }

    // Get information about the file
    DWORD attr = GetFileAttributes(fullFilePath.c_str());
    if (attr == 0xFFFFFFFF)
    {
    	request.Socket.SendContent("file not found");
    	return;
    }

    // Directory handle: check for default files
    if (attr & FILE_ATTRIBUTE_DIRECTORY)
    {
    	if (FileExists(fullFilePath + "\\index.html"))
        {
        	fullFilePath += "\\index.html";
            attr = GetFileAttributes(fullFilePath.c_str());
        }
        else if (FileExists(fullFilePath + "\\index.htm"))
        {
        	fullFilePath += "\\index.htm";
            attr = GetFileAttributes(fullFilePath.c_str());
        }
    }

    // Directory handle: show directory contents
    if (attr & FILE_ATTRIBUTE_DIRECTORY)
    {
    	if (!GAllowDirectoryListing)
        {
        	request.Socket.SendContent("no directory listing allowed");
            return;
        }
    	AnsiString code = "<html><head><title>";
        code += filePart;
        code += "</title></head><body><h1>Directory contents for ";
        code += HTMLEscape(filePart);
        code += "</h1><hr><ul>";
        WIN32_FIND_DATA fd;
        HANDLE h = FindFirstFile((fullFilePath + "\\*.*").c_str(), &fd);
        if (h != INVALID_HANDLE_VALUE)
        {
        	do
            {
            	if ((fd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) ||
                	(fd.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM))
                {
                	continue;
                }
            	code += "<li><a href=\"";
			    GetFullPathName((fullFilePath + "\\" + AnsiString(fd.cFileName)).c_str(), MAX_PATH, fullPath, 0);
                code += URLEscape(fd.cFileName);
                if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                {
                	code += '/';
                }
                code += "\">" + HTMLEscape(fd.cFileName) + "</a> (";
				if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                {
                	code += "directory)</li>";
                }
                else
                {
	                unsigned __int64 size = 0;
	                memcpy(&size, &fd.nFileSizeLow, 4);
	                memcpy(((char*)&size)+4, &fd.nFileSizeHigh, 4);
	                char tmp[100];
	                _ui64toa(size, tmp, 10);
	                code += AnsiString(tmp) + " bytes)</li>";
                }
            }
            while (FindNextFile(h, &fd));
        	FindClose(h);
        }
        else
        {
        	code += "<li><i>Error, cannot list files</i></li>";
        }
        code += "</ul><hr>Served at " + DateTimeToStr(Now());
        code += "<br></body></html>";
        request.Socket.SendContent(code.c_str());
        return;
    }

    // Normal (non-hidden, non-system) file, send it
    if (!(	(attr & FILE_ATTRIBUTE_HIDDEN) ||
    		(attr & FILE_ATTRIBUTE_SYSTEM)))
    {
    	bool sendFile = true;

    	// Check if the file is a content DLL
        if (GAllowContentDLLs &&
        	fullFilePath.Length() > 4 &&
            fullFilePath.SubString(fullFilePath.Length() - 3, MAX_PATH) == ".dll")
        {
        	// Try to load the DLL
            HINSTANCE dll = LoadLibrary(fullFilePath.c_str());
            if (dll)
            {
            	FARPROC proc = GetProcAddress(dll, "content");
                if (!proc) proc = GetProcAddress(dll, "_content");
                if (proc)
                {
					typedef int (_stdcall * FSend)(SOCKET, const char*, int);
                	typedef void _stdcall (*FContent)(const char*, SOCKET, FSend);
                    FContent content = (FContent)proc;
                    chdir(ExtractFileDir(fullFilePath).c_str());
                    content(request.Address, request.Socket.GetSocket(), DLLSend);
                    sendFile = false;
                }
                FreeLibrary(dll);
            }
        }

        // We should send the file
        if (sendFile)
        {
	    	request.Socket.SendFile(fullFilePath.c_str());
        }
    }

    // Update
    Main->sbStatus->Panels->Items[1]->Text = "Served " + AnsiString(request.Address);
}

static void SaveOptions()
{
	TRegistry* reg = new TRegistry();
    reg->LazyWrite = false;
    if (reg->CreateKey("\\Software\\Runtime Legend") &&
    	reg->CreateKey("\\Software\\Runtime Legend\\BCB Web Server") &&
        reg->OpenKey("\\Software\\Runtime Legend\\BCB Web Server", false))
    {
    	reg->WriteString("Content Root", GBasePath);
        reg->WriteBool("Allow Directory Listing", GAllowDirectoryListing);
        reg->WriteBool("Start Automatically", GAutoStart);
        reg->WriteBool("Hide Main Window", GHideMainWindow);
        reg->WriteBool("Allow Content DLLs", GAllowContentDLLs);
    	reg->CloseKey();
    }
    delete reg;
}

static void LoadOptions()
{
	TRegistry* reg = new TRegistry();
    reg->LazyWrite = false;
    if (reg->OpenKey("\\Software\\Runtime Legend\\BCB Web Server", false))
    {
    	try
        {
	    	AnsiString value;
        	value = reg->ReadString("Content Root");
            GBasePath = value;
        }
        catch(...){}
    	try
        {
	    	bool value;
        	value = reg->ReadBool("Allow Directory Listing");
            GAllowDirectoryListing = value;
        }
        catch(...){}
    	try
        {
	    	bool value;
        	value = reg->ReadBool("Start Automatically");
            GAutoStart = value;
        }
        catch(...){}
    	try
        {
	    	bool value;
        	value = reg->ReadBool("Hide Main Window");
            GHideMainWindow = value;
        }
        catch(...){}
    	try
        {
	    	bool value;
        	value = reg->ReadBool("Allow Content DLLs");
            GAllowContentDLLs = value;
        }
        catch(...){}
    	reg->CloseKey();
    }
    delete reg;
}
//---------------------------------------------------------------------------
void __fastcall TMain::OnTrayMessage(TMessage msg)
{
	if (msg.LParam == WM_LBUTTONDBLCLK)
    {
    	mTrayShowWindowClick(0);
    }
	else if (msg.LParam == WM_RBUTTONDOWN || msg.LParam == WM_CONTEXTMENU)
    {
    	POINT p;
    	GetCursorPos(&p);
    	mTray->Popup(p.x, p.y);
    }
}
//---------------------------------------------------------------------------
__fastcall TMain::TMain(TComponent* Owner)
	: TForm(Owner)
{
	char directory[MAXDIR];
    getcurdir(0, directory);
    GBasePath = AnsiString((char)('A' + getdisk())) + ":\\" + directory;
    LoadOptions();
	CWebServer::Initialize();
    server.SetCallback(Callback);
}
//---------------------------------------------------------------------------
void __fastcall TMain::FormClose(TObject *Sender, TCloseAction &Action)
{
	server.Stop();
    SaveOptions();
}
//---------------------------------------------------------------------------
void __fastcall TMain::Timer1Timer(TObject *Sender)
{
	static bool first = true;
    if (first)
    {
    	first = false;
	    if (GHideMainWindow) mServerToTrayClick(Sender);
    }
	server.Wait(0);
}
//---------------------------------------------------------------------------
void __fastcall TMain::mHelpAboutClick(TObject *Sender)
{
	Application->CreateForm(__classid(TAboutBox), &AboutBox);
	AboutBox->ShowModal();
    delete AboutBox;
}
//---------------------------------------------------------------------------
void __fastcall TMain::mServerTurnOnClick(TObject *Sender)
{
	if (!server.Listen())
    {
    	Application->MessageBox("Cannot start the server", "Error", MB_ICONERROR);
    	return;
    }
    Timer1->Enabled = true;
    sbStatus->Panels->Items[0]->Text = "Listening on port 80";
    mServerTurnOn->Enabled = false;
    mServerTurnOff->Enabled = true;
    mToolsLaunchBrowser->Enabled = true;
    sbTurnOn->Enabled = false;
    sbTurnOff->Enabled = true;
    sbLaunchBrowser->Enabled = true;
    mTrayTurnOn->Enabled = false;
    mTrayTurnOff->Enabled = true;
}
//---------------------------------------------------------------------------
void __fastcall TMain::mServerTurnOffClick(TObject *Sender)
{
    Timer1->Enabled = false;
    sbStatus->Panels->Items[0]->Text = "The server is turned off";
    mServerTurnOn->Enabled = true;
    mServerTurnOff->Enabled = false;
    mToolsLaunchBrowser->Enabled = false;
    sbTurnOn->Enabled = true;
    sbTurnOff->Enabled = false;
    sbLaunchBrowser->Enabled = false;
    mTrayTurnOn->Enabled = true;
    mTrayTurnOff->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TMain::sbExitClick(TObject *Sender)
{
	mServerExitClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMain::mServerExitClick(TObject *Sender)
{
	GCanExit = true;
	Close();	
}
//---------------------------------------------------------------------------
void __fastcall TMain::sbTurnOnClick(TObject *Sender)
{
	mServerTurnOnClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMain::sbTurnOffClick(TObject *Sender)
{
	mServerTurnOffClick(Sender);	
}
//---------------------------------------------------------------------------
void __fastcall TMain::mToolsViewFilesinExplorerClick(TObject *Sender)
{
	ShellExecute(Handle, 0, ".", 0, 0, SW_SHOW);  	
}
//---------------------------------------------------------------------------
void __fastcall TMain::mToolsLaunchBrowserClick(TObject *Sender)
{
	ShellExecute(Handle, 0, "http://localhost/", 0, 0, SW_SHOW);

}
//---------------------------------------------------------------------------
void __fastcall TMain::SpeedButton3Click(TObject *Sender)
{
	mHelpAboutClick(Sender);	
}
//---------------------------------------------------------------------------
void __fastcall TMain::sbLaunchBrowserClick(TObject *Sender)
{
	mToolsLaunchBrowserClick(Sender);	
}
//---------------------------------------------------------------------------
void __fastcall TMain::sbExploreClick(TObject *Sender)
{
	mToolsViewFilesinExplorerClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMain::mToolsOptionsClick(TObject *Sender)
{
	Application->CreateForm(__classid(TOptionsBox), &OptionsBox);
    OptionsBox->edRootDir->Text = GBasePath;
    OptionsBox->cbAllowListing->Checked = GAllowDirectoryListing;
    OptionsBox->cbStartAutomatically->Checked = GAutoStart;
    OptionsBox->cbHideMainWindow->Checked = GHideMainWindow;
    OptionsBox->cbAllowContentDLLs->Checked = GAllowContentDLLs;
    if (OptionsBox->ShowModal() == mrOk)
    {
    	GBasePath = OptionsBox->edRootDir->Text;
        GAllowDirectoryListing = OptionsBox->cbAllowListing->Checked;
    	GAutoStart = OptionsBox->cbStartAutomatically->Checked;
    	GHideMainWindow = OptionsBox->cbHideMainWindow->Checked;
	    GAllowContentDLLs = OptionsBox->cbAllowContentDLLs->Checked;
        SaveOptions();
    }
    delete OptionsBox;
}
//---------------------------------------------------------------------------
void __fastcall TMain::SpeedButton2Click(TObject *Sender)
{
	mToolsOptionsClick(Sender);	
}
//---------------------------------------------------------------------------
void __fastcall TMain::FormDestroy(TObject *Sender)
{
	NOTIFYICONDATA nid;
    nid.cbSize = sizeof(nid);
    nid.hWnd = Handle;
    nid.uID = 100;
    nid.uFlags = 0;
    Shell_NotifyIcon(NIM_DELETE, &nid);
}
//---------------------------------------------------------------------------
void __fastcall TMain::FormCreate(TObject *Sender)
{
	NOTIFYICONDATA nid;
    nid.cbSize = sizeof(nid);
    nid.hWnd = Handle;
    nid.uID = 100;
    nid.uFlags = NIF_ICON|NIF_MESSAGE|NIF_TIP;
    nid.uCallbackMessage = WM_USER + 100;
    nid.hIcon = Application->Icon->Handle;
    strcpy(nid.szTip, "Web Server");
    Shell_NotifyIcon(NIM_ADD, &nid);
    if (GAutoStart) mServerTurnOnClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMain::mTrayTurnOnClick(TObject *Sender)
{
	mServerTurnOnClick(Sender);	
}
//---------------------------------------------------------------------------
void __fastcall TMain::mTrayTurnOffClick(TObject *Sender)
{
	mServerTurnOffClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMain::mTrayOptionsClick(TObject *Sender)
{
	mToolsOptionsClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMain::mTrayAboutClick(TObject *Sender)
{
	mHelpAboutClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMain::mTrayExitClick(TObject *Sender)
{
	mServerExitClick(Sender);	
}
//---------------------------------------------------------------------------
void __fastcall TMain::mTrayShowWindowClick(TObject *Sender)
{
	if (Visible)
    {
    	Application->BringToFront();
    	return;
    }
	SetWindowLong(Application->Handle, GWL_EXSTYLE,
    	GetWindowLong(Application->Handle, GWL_EXSTYLE)&~WS_EX_TOOLWINDOW);
	ShowWindow(Application->Handle, SW_SHOW);
    Show();
}
//---------------------------------------------------------------------------
void __fastcall TMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	if (GCanExit)
    {
    	CanClose = true;
    }
    else
    {
		ShowWindow(Application->Handle, SW_HIDE);
		SetWindowLong(Application->Handle, GWL_EXSTYLE,
	    	GetWindowLong(Application->Handle, GWL_EXSTYLE)|WS_EX_TOOLWINDOW);
	    Hide();
		CanClose = false;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMain::mServerToTrayClick(TObject *Sender)
{
	Close();	
}
//---------------------------------------------------------------------------
void __fastcall TMain::sbToTrayClick(TObject *Sender)
{
	mServerToTrayClick(Sender);	
}
//---------------------------------------------------------------------------
