// ServerMe.H Test program

// Make MSVC stop whining when using portable stuff
#define _CRT_SECURE_NO_WARNINGS

#include "ServerMe.H"
#include <signal.h>
#include <string>

using namespace std;

static bool running = true;     // Running flag

// Called for every connection
static void callback(SWebServerRequest& request, void*)
{
    // Form a response
    string code;
    code = "<html><head><title>Response</title></head><body><h1>Test</h1>";
    if (strstr(request.Address, "/bye"))
    {
        code += "<b>Byee!</b>";
        running = false;
    }
    else
    {
        code += "<b>Hello there!</b> You asked for " + string(request.Address);
        code += ", but this server doesn't really support such requests.<br><br>";
        code += "Ask for <a href='/bye'>bye</a> to quit.";
        code += "<br><br>Better luck next time! Bye...</body></html>";
    }
    request.Socket.SendContent(code.c_str());
}

// Called when the program is asked to terminate
static void termsig(int signal)
{
    if (signal == SIGTERM)
    {
        running = false;
    }
}

int main()
{
    // Initialize the server
    if (!CWebServer::Initialize())
    {
        fprintf(stderr, "Failed to initialize\n");
        return EXIT_FAILURE;
    }

    // Start listening for connections
    CWebServer server;
    if (!server.Listen(8080))
    {
        fprintf(stderr, "Failed to start listening for connections\n");
        return EXIT_FAILURE;
    }

    signal(SIGTERM, termsig);

    // Setup the callback
    server.SetCallback(callback);

    // Handle requests
    while (running)
    {
        // Block for 1sec until a request comes (the 1sec delay
        // allows us to periodically check if we should terminate)
        server.Wait(1000);
    }

    // Stop the server
    server.Stop();

    // Shutdown the server
    CWebServer::Shutdown();

    return 0;
}
